"""
Seed functions are the functions that produce the first sequence of the pipeline
of sequence transformations.

They're not the outer most layer:
The outer-most layer contains the whole pipeline,
and has as arguments any parameters that the steps of the pipeline might need.

A seed functions is called repeatedly to produce either an item at a time (`n=1`),
a chunk (`n>1`), or an iterable (`n=0`) that may or may not be a finite one. 
"""
from dataclasses import dataclass
from functools import partial
from typing import Callable, Union, Collection, Iterable, Tuple, Any, Dict, Sequence
from random import randint
import random
import itertools
import string
import abc


class RandomGeneratorBase:
    chunk_container: Callable = list

    @abc.abstractmethod
    def generate_one(self):
        raise NotImplemented('Need to provide a generate_one method')

    def __call__(self, n=1):
        if n == 1:
            return self.generate_one()
        else:
            return self.chunk_container(self.read(n))

    def __iter__(self):
        while True:
            yield self()

    def read(self, n=1):
        return itertools.islice(self, n)


@dataclass
class RandomGenerator(RandomGeneratorBase):
    random_gen: Callable[[], Any]
    chunk_container: Callable = list

    def generate_one(self):
        return self.random_gen()


@dataclass
class RandomCategoricalGenerator(RandomGeneratorBase):
    """Generate categorical data in a controlled random way

    # TODO: See why this runs forever!!!
    # >>> r = RandomCategoricalGenerator()
    # >>> it = r(None)
    # >>> assert r() in r.categories # True or False
    # >>> assert r(n=1) in r.categories # same as r()
    # >>> result = r(n=3)
    # >>> assert isinstance(result, list)  # e.g. [False, True, False]
    # >>> r = RandomCategoricalGenerator(chunk_container=tuple)
    # >>> result = r(n=3)
    # >>> assert isinstance(result, tuple)  # e.g. (False, True, False)
    # >>> result = itertools.islice(r(None), 0, 4)
    # >>> assert set(result).issubset(r.categories)
    """

    categories: Sequence = (True, False)
    chunk_container: Callable = list

    # TODO: Allow possibility of weighted categories (p argument of np.random.choice)
    # TODO: Being able to specify random seed

    def __post_init__(self):
        self.categories = list(self.categories)  # make it sized
        self.n_categories = len(self.categories)

    def generate_one(self):
        return random.choice(self.categories)


def rand_string(str_size=(1, 9), alphabet=string.ascii_letters, joiner=''.join):
    if isinstance(str_size, (float, int)):
        str_size = (int(str_size), int(str_size))
    word_size = randint(*str_size)
    return joiner([random.choice(alphabet) for _ in range(word_size)])


@dataclass
class RandomStringGenerator(RandomGeneratorBase):
    """Random string generator

    >>> gen = RandomStringGenerator()
    >>> t = gen()
    >>> assert isinstance(t, str)
    >>> t  # doctest: +SKIP
    'CmwRB'
    >>> t = RandomStringGenerator(3, 'ab')(n=4)
    >>> t  # doctest: +SKIP
    ['bbb', 'bba', 'bab', 'abb']
    >>> assert all(isinstance(x, str) for x in t)
    >>> len(t)
    4
    """

    str_size: Union[int, Tuple[int, int]] = (1, 9)
    alphabet: Collection = string.ascii_letters
    joiner: Callable[[Iterable], str] = ''.join

    def __post_init__(self):
        if isinstance(self.str_size, (float, int)):
            self.str_size = (int(self.str_size), int(self.str_size))

    def generate_one(self):
        return rand_string(self.str_size, self.alphabet, self.joiner)


# TODO: Meant for postelizing random generators. Specify only a type, or enum etc.
DFLT_RAND_FUNC_FOR_TYPE = (
    (int, partial(random.randint, 0, 100)),
    (float, partial(random.uniform, 0, 100)),
    (str, partial(rand_string)),
)


ArglessFunc = Callable[[], Any]
GenForField = Union[Dict[str, ArglessFunc], Iterable[Tuple[str, ArglessFunc]]]


@dataclass
class RandomDictGenerator(RandomGeneratorBase):
    """Fixed-schema dict generator.

    Note: More power found in `slink` `IterativeDictProcessing` and `generate_dict`

    `RandomDictGenerator` is left here because it has the common `RandomGeneratorBase`
    interface.

    >>> from functools import partial
    >>> rand_gen = RandomDictGenerator(gen_for_field=(
    ... ('rpm', partial(random.uniform, 100, 1000)),
    ... ('temperature', partial(random.randint, 15, 25)),
    ... ('id', partial(rand_string, str_size=(2, 5), alphabet='0123456789abcdef')),
    ... ('kind', RandomCategoricalGenerator(categories=[True, False, None]))
    ... ))
    >>> t = rand_gen()
    >>> t  # doctest: +SKIP
    {'rpm': 172.58455409094074, 'temperature': 21, 'id': '51a14', 'kind': True}
    >>> assert isinstance(t, dict)
    >>> assert list(t) == ['rpm', 'temperature', 'id', 'kind']

    You can also ask for several at once, by default, returned in a list.

    >>> rand_gen(3)  # doctest: +SKIP
    [{'rpm': 275.35599992556627, 'temperature': 21, 'id': '509', 'kind': True},
    {'rpm': 378.8770605317944, 'temperature': 23, 'id': '5657a', 'kind': None},
    {'rpm': 826.381460898361, 'temperature': 23, 'id': '4a53d', 'kind': False}]

    rand_gen is also an (infinite) iterable:

    >>> for x in rand_gen:
    ...     break
    >>> assert list(x) == ['rpm', 'temperature', 'id', 'kind']

    You can ask for an iterator with a limited number of elements.

    >>> it = rand_gen.read(3)
    >>> import itertools
    >>> assert isinstance(it, itertools.islice)
    >>> y = list(rand_gen.read(3))
    >>> assert len(y) == 3
    >>> assert list(y[0]) == ['rpm', 'temperature', 'id', 'kind']
    """

    gen_for_field: GenForField = (
        ('rpm', partial(random.uniform, 100, 1000)),
        ('temperature', partial(random.randint, 15, 25)),
        ('id', partial(rand_string, str_size=(2, 5), alphabet='0123456789abcdef')),
        ('kind', RandomCategoricalGenerator(categories=[True, False, None])),
    )
    chunk_container: Callable = list

    def __post_init__(self):
        self.gen_for_field = dict(self.gen_for_field)

    def generate_one(self):
        return {k: v() for k, v in self.gen_for_field.items()}


class RandomTupleGenerator(RandomDictGenerator):
    """Generate fixed-schema tuples (of numericals or tuples)
    Can be used to generate other fixed-schema objects (dict, custom classes, etc.)
    """



"""Tools to make sequences

Highlights:

>>> from slink.sequences import IterativeDictProcessing
>>> f = IterativeDictProcessing(
...     phase=lambda session: session * 10,
...     something_dependent=lambda session, phase: session + phase,
...     something_independent=lambda: 'hi'
... )
>>> f({'session': 2})
{'session': 2, 'phase': 20, 'something_dependent': 22, 'something_independent': 'hi'}

>>> from slink.sequences import dict_generator, Repeater
>>> import itertools
>>> counter = itertools.count()
>>> f = dict_generator(dict(
...     x=7,  # will be replaced with ReturnObj(y), an equivalent of lambda: 7
...     _1=Repeater(3),
...     y=lambda: next(counter),  # will return 0, then 1, then 2,...
...     z=lambda x, y: x * y),
...     1
... )
>>> list(f())
[{'x': 7, 'y': 0, 'z': 0}, {'x': 7, 'y': 1, 'z': 7}, {'x': 7, 'y': 2, 'z': 14}]

"""

from itertools import accumulate
from functools import partial
from typing import Callable, Any, Iterator
import random
import copy
from dataclasses import dataclass

from i2 import MultiFunc, Sig


def dict_generator(*formulas):
    """helper function to make DictChain objects.

    >>> import itertools
    >>> counter = itertools.count()
    >>> f = dict_generator(dict(
    ...     x=7,  # will be replaced with ReturnObj(y), an equivalent of lambda: 7
    ...     _1=Repeater(3),
    ...     y=lambda: next(counter),  # will return 0, then 1, then 2,...
    ...     z=lambda x, y: x * y),
    ...     1
    ... )
    >>> list(f())
    [{'x': 7, 'y': 0, 'z': 0}, {'x': 7, 'y': 1, 'z': 7}, {'x': 7, 'y': 2, 'z': 14}]
    >>>
    >>> counter = itertools.count()
    >>> f = dict_generator(dict(
    ...     x=7,  # will be replaced with ReturnObj(y), an equivalent of lambda: 7
    ...     _1=Repeater(3),
    ...     y=lambda: next(counter),  # will return 3, then 4, then 5,...
    ...     z=lambda x, y: x * y),
    ...     2  # will be replaced with Repeater(2)
    ... )
    >>> t = list(f())
    >>> t  # doctest: +NORMALIZE_WHITESPACE
    [{'x': 7, 'y': 0, 'z': 0}, {'x': 7, 'y': 0, 'z': 0},
    {'x': 7, 'y': 1, 'z': 7}, {'x': 7, 'y': 1, 'z': 7},
    {'x': 7, 'y': 2, 'z': 14}, {'x': 7, 'y': 2, 'z': 14}]

    Equivalently,

    >>> counter = itertools.count()
    >>> f = dict_generator(
    ...     dict(x=7),
    ...     3,  # will be replaced with Repeater(3)
    ...     dict(y=lambda: next(counter), z=lambda x, y: x * y),
    ...     2  # will be replaced with Repeater(2)
    ... )
    >>> assert list(f()) == t
    """
    return DictChain(**dict(_prepare_formulas(formulas)))


# TODO: Shares big part with IterativeDictProcessing. Should we merge?
class DictChain(MultiFunc):
    """Make objects that generate schemaed and formulaed dicts with repetition

    >>> import itertools
    >>>
    >>> counter = itertools.count()
    >>> def count():
    ...     return next(counter)
    ...
    >>> f = DictChain(
    ...     x=lambda: 7,
    ...     _1=Repeater(3),
    ... )
    >>> list(f())
    [{'x': 7}, {'x': 7}, {'x': 7}]
    >>>
    >>> f = DictChain(
    ...     x=lambda: 7,
    ...     _1=Repeater(3),
    ...     y=lambda: next(counter),  # will return 0, then 1, then 2,...
    ... )
    >>> list(f())
    [{'x': 7, 'y': 0}, {'x': 7, 'y': 1}, {'x': 7, 'y': 2}]
    >>>
    >>>
    >>> f = DictChain(
    ...     x=lambda: 7,
    ...     _1=Repeater(3),
    ...     y=lambda: next(counter),  # will return 3, then 4, then 5,....
    ...     z=lambda x, y: x * y,
    ... )
    >>> list(f())
    [{'x': 7, 'y': 3, 'z': 21}, {'x': 7, 'y': 4, 'z': 28}, {'x': 7, 'y': 5, 'z': 35}]
    >>>
    >>>
    >>> f = DictChain(
    ...     x=lambda: 7,
    ...     _1=Repeater(3),
    ...     y=lambda: next(counter),  # will return 6, then 7, then 8,...
    ...     z=lambda x, y: x * y,
    ...     _2=Repeater(2)
    ... )
    >>> list(f())  # doctest: +NORMALIZE_WHITESPACE
    [{'x': 7, 'y': 6, 'z': 42}, {'x': 7, 'y': 6, 'z': 42},
    {'x': 7, 'y': 7, 'z': 49}, {'x': 7, 'y': 7, 'z': 49},
    {'x': 7, 'y': 8, 'z': 56}, {'x': 7, 'y': 8, 'z': 56}]
    """

    def __init__(self, **unnamed_funcs):
        super().__init__(**unnamed_funcs)
        self.sigs = {name: Sig(func) for name, func in self.items()}

    def __call__(self, seed_dict=None, preproc=copy.copy):
        # return map(next, self._all_calls(seed_dict, preproc))
        return flatten_generators_recursively(self._call_(seed_dict, preproc))

    def _call_(self, seed_dict=None, preproc=copy.copy):
        if seed_dict is None:
            seed_dict = dict()
        elif callable(seed_dict):
            seed_dict_factory = seed_dict
            seed_dict = seed_dict_factory()
        if preproc:  # should we
            seed_dict = preproc(seed_dict)
        # assert isinstance(seed_dict, dict)

        # get the first (name, func) pair
        if len(self) > 0:
            (name, func), *remaining = self.items()
            next_repeater = DictChain(**dict(remaining))
            if isinstance(func, Repeater):
                repeater = func
                yield from map(next_repeater, repeater(seed_dict, seed_dict))
            else:
                seed_dict[name] = _call_from_dict(seed_dict, func, self.sigs[name])
                yield from next_repeater(seed_dict, preproc)
        else:
            yield seed_dict


def _call_from_dict(kwargs: dict, func: Callable, sig: Sig):
    """A i2.call_forgivingly optimized for our purpose

    The sig argument needs to be the Sig(func) to work correctly.
    """
    args, kwargs = sig.args_and_kwargs_from_kwargs(
        kwargs,
        allow_excess=True,
        ignore_kind=True,
        allow_partial=False,
        apply_defaults=True,
    )
    return func(*args, **kwargs)


class IterativeDictProcessing(MultiFunc):
    """Generate or transform a dict iteratively from a set of iterative rules (functions)

    This is useful when you need to generate dicts (or any other vector --
    i.e fixed-width schema-ed tuple/list/etc) that have correlated values.

    Deterministic example with specified input:

    >>> f = IterativeDictProcessing(
    ...     phase=lambda session: session * 10,
    ...     something_dependent=lambda session, phase: session + phase,
    ...     something_independent=lambda: 'hi'
    ... )
    >>> f({'session': 2})
    {'session': 2, 'phase': 20, 'something_dependent': 22, 'something_independent': 'hi'}

    Non-deterministic example with empty input.
    Here we demonstrate that the original input that seeds the iterative generation
    can be given by a function that will generate a random input.

    >>> import functools, random
    >>> f = IterativeDictProcessing(
    ...     session=functools.partial(random.uniform, 2, 9),
    ...     block=lambda session: session * 10,
    ...     something_dependent=lambda session, block: session + block,
    ...     something_independent=lambda: random.choice([True, False, None])
    ... )
    >>> # f() is equivalent to specifying empty dict; f({}); or factory; f(dict)
    >>> f()  # doctest: +SKIP
    {'session': 8.673108499155791, 'block': 86.7310849915579,
    'something_dependent': 95.4041934907137, 'something_independent': False}

    Note that, akin to working with a spreadsheet, sometimes you need to create
    intermediate variables to use in your formulas, but don't want those fields in
    your final data.
    To acheive this just filter out the unwanted fields.
    You can use `slink.util.select_fields` for this.

    """

    def __init__(self, **named_funcs):
        super().__init__(**named_funcs)
        self.sigs = {name: Sig.sig_or_default(func) for name, func in self.items()}

    def __call__(self, seed_dict=None, preproc=copy.copy):
        if seed_dict is None:
            seed_dict = dict()
        elif callable(seed_dict):
            seed_dict_factory = seed_dict
            seed_dict = seed_dict_factory()
        if preproc:  # should we
            seed_dict = preproc(seed_dict)
        assert isinstance(seed_dict, dict)
        for assign_to_name, func in self.items():
            seed_dict[assign_to_name] = _call_from_dict(
                seed_dict, func, self.sigs[assign_to_name]
            )
        return seed_dict


import itertools
from typing import Generator
from typing import NewType, Tuple, Callable, Union


def new_type(name, typ, doc=None):
    t = NewType(name, type)
    if doc:
        t.__doc__ = doc
    return t


@dataclass
class Repeater:
    """Helper class to define the repetition of an object based on a fixed number or
    a number (unknown at the time of instantiation) that will be defined in a
    seed_dict"""

    n_repetitions: Union[int, str]

    def __call__(self, obj, seed_dict=None):
        if isinstance(self.n_repetitions, str):
            # if n_repetitions is a string, it's a field; get the number_of_repetitions
            n_repetitions = (seed_dict or {}).get(self.n_repetitions, None)
            if n_repetitions is None:
                raise KeyError(f'{self.n_repetitions} key not found in {seed_dict}')
        else:
            n_repetitions = self.n_repetitions
        return itertools.repeat(obj, n_repetitions)


# TODO: Integrate these types and make DictChain more explicit in it's interface,
#   leaving the dict_generator have more flexible argument handling.
# TODO: Formulas is almost what IterativeDictProcessing is now. Merge.
Identifier = new_type('Identifier', str)  # str.isidentifier
Formula = new_type('Formula', Tuple[Identifier, Callable])
Formulas = new_type('Formulas', Tuple[Formula])
FormulaSpec = new_type('FormulaSpec', Union[Formula, dict])
RepeaterSpec = new_type('RepeaterSpec', Union[Repeater, int, str])
Step = new_type('Step', Union[Formulas, Repeater])


def flatten_generators_recursively(x: Generator):
    for xi in x:
        if isinstance(xi, Generator):
            yield from flatten_generators_recursively(xi)
        else:
            yield xi


@dataclass
class ReturnObj:
    obj: Any

    def __call__(self):
        return self.obj


def _prepare_formula_dict(formula_dict):
    for k, v in formula_dict.items():
        if not isinstance(v, Callable):
            yield k, ReturnObj(v)
        else:
            yield k, v


def _prepare_formulas(formulas):
    for i, formula in enumerate(formulas):
        if isinstance(formula, dict):
            yield from _prepare_formula_dict(formula)
        elif isinstance(formula, Repeater):
            yield f'repeater_{i}', formula
        elif isinstance(formula, (int, str)):
            yield f'repeater_{i}', Repeater(formula)
        else:
            raise TypeError(
                f'formulas should be dicts, Repeaters, ints or strings: ' f'{formula}'
            )


def call_repeatedly(func, *args, **kwargs) -> Iterator:
    """
    >>> func = enumerate(range(4)).__next__
    >>> iterator = call_repeatedly(func)
    >>> list(iterator)
    [(0, 0), (1, 1), (2, 2), (3, 3)]
    """
    return iter(partial(func, *args, **kwargs), object())


def mk_monotone_sequence(delta_val_func=random.random, *args, start=0, **kwargs):
    """Make a monotone sequence of numbers by accumulating random time durations/deltas.

    >>> from lined import Pipe
    >>> import itertools
    >>> get_slice = Pipe(itertools.islice, list)
    >>>
    >>> it = mk_monotone_sequence()
    >>> t = get_slice(it, 4)
    >>> assert len(t) == 4
    >>> assert t[0] == 0
    >>> t  # doctest: +SKIP
    [0, 0.522353704156762, 0.9681615026643958, 1.0927423253813298]
    >>> it = mk_monotone_sequence(random.uniform, 10, 100, start=7)
    >>> t = get_slice(it, 3)
    >>> assert len(t) == 3
    >>> assert t[0] == 7
    >>> t  # doctest: +SKIP
    [7, 62.808676729760556, 129.67231010126588]
    """
    return accumulate(call_repeatedly(delta_val_func, *args, **kwargs), initial=start)



"""Make timestamped data"""
from functools import partial
from creek import Creek
from lined import Line, LineParametrized, iterize


# ---------------------------------------------------------------------------------------
# Using creek to make a stream controlled by categoricals
"""
 a, b...  ──▶  wf_for(a), wf_for(b)...
"""
cat_map = {'a': [1, 2, 3], 'b': [4, 5, 6]}
src = 'aabbaab'


class CatCreek(Creek):
    def __init__(self, stream, cat_map):
        super().__init__(stream)
        self.cat_map = cat_map

    def data_to_obj(self, item):
        return self.cat_map[item]


class test_cat_creek:
    stream = CatCreek(stream=src, cat_map=cat_map)

    assert list(stream) == [
        [1, 2, 3],
        [1, 2, 3],
        [4, 5, 6],
        [4, 5, 6],
        [1, 2, 3],
        [1, 2, 3],
        [4, 5, 6],
    ]


# ---------------------------------------------------------------------------------------
# Composition of creeks using lined
"""
              ┌────────────┐
 cat_map  ──▶ │ TwCatCreek │ ──▶  output
              └────────────┘
                ▲
                │ stream
                │
              ┌────────────┐
 stream   ──▶ │  IdCreek   │
              └────────────┘
"""

# TW: Equivalent to just Creek
class IdCreek(Creek):
    def data_to_obj(self, item):
        return item


from lined import LineParametrized

pipe = LineParametrized(IdCreek, CatCreek)

assert list(pipe(src, cat_map=cat_map)) == [
    [1, 2, 3],
    [1, 2, 3],
    [4, 5, 6],
    [4, 5, 6],
    [1, 2, 3],
    [1, 2, 3],
    [4, 5, 6],
]


# ---------------------------------------------------------------------------------------
# analogue of iterize for creek

"""
 a, b...  ──▶  wf_for(a), wf_for(b)...  ──▶  transf_wf, transf_wf_b...
"""


# %%


def creek_iterize(f, stream):
    m = Creek(stream)
    m.data_to_obj = f
    return m


def func_to_creek(func):
    return partial(creek_iterize, func)


def test_func_to_creek():
    stream = [1, 2, 3]

    def double(item):
        return 2 * item

    assert list(func_to_creek(double)(stream)) == [2, 4, 6]


test_func_to_creek()

# ---------------------------------------------------------------------------------------
# Transform mapping into a function


def dict_to_func(d):
    return d.__getitem__


g = dict_to_func(cat_map)
assert g('a') == [1, 2, 3]


def dict_as_param(func):
    def g(d):
        return func(**d)

    return g


def test_dict_as_param():
    func = lambda x, y: x + y
    f = dict_as_param(func)
    assert f(dict(x=2, y=3)) == 5
    # but note that f isn't pickle-able


test_dict_as_param()


def test_func_to_creek_and_dict_to_func(cat_map=cat_map, src=src):
    stream = func_to_creek(dict_to_func(cat_map))(src)
    assert list(stream) == [
        [1, 2, 3],
        [1, 2, 3],
        [4, 5, 6],
        [4, 5, 6],
        [1, 2, 3],
        [1, 2, 3],
        [4, 5, 6],
    ]


test_func_to_creek_and_dict_to_func()


# TW: I stopped here
# ---------------------------------------------------------------------------------------
# Rewriting Christian's functions with creek

# %%

# Christian's initial function
def annot_timestamping(n_annot, start, end, wiggle=100):
    """
    Make synthetic timestamps
    """
    base = list(np.linspace(start, end, n_annot))
    wiggled = (
        base[0:1]
        + [i + np.random.randint(-wiggle, wiggle) for i in base[1:-1]]
        + base[-1:]
    )
    return wiggled


cat_lin_map = {
    'a': dict(start=1, stop=25, num=12),
    'b': dict(start=0, stop=5, num=30),
}

# %%

src = 'aabbaab'
stream = Line(dict_to_func, func_to_creek)(cat_lin_map)(src)

# %%

list(stream)

# %%

import numpy as np

p = Line(dict_as_param, func_to_creek)(np.linspace)

# %%

list(p(stream))[:3]

# %%

dict_as_param(np.linspace)({'start': 1, 'stop': 25, 'num': 12})


# wiggle an array
def wiggle(arr, wig_factor):
    wiggling = np.random.randint(-wig_factor, wig_factor, len(arr))
    wiggling[0] = 0
    wiggling[-1] = 0

    return arr + wiggling


def wiggle_by(wig_factor):
    return partial(wiggle, wig_factor=wig_factor)


# ---------------------------------------------------------------------------------------
# package everything using lined


# symbols -> Iterable[dict] -> Iterable[np.array] - --(wiggle each array) --> Iterable[
#     np.array]


# This replaces the function annot_timestamping
src = 'aabbaab'
p1 = Line(dict_to_func, func_to_creek)(cat_lin_map)
stream = p1(src)
p2 = Line(dict_as_param, func_to_creek)(np.linspace)
stream_final = p2(stream)
list(stream_final)[:3]

# %%



"""Slink: Sequence generator

Highlights:

>>> from slink import IterativeDictProcessing
>>> f = IterativeDictProcessing(
...     phase=lambda session: session * 10,
...     something_dependent=lambda session, phase: session + phase,
...     something_independent=lambda: 'hi'
... )
>>> f({'session': 2})
{'session': 2, 'phase': 20, 'something_dependent': 22, 'something_independent': 'hi'}

>>> from slink import dict_generator, Repeater
>>> import itertools
>>> counter = itertools.count()
>>> f = dict_generator(dict(
...     x=7,  # will be replaced with ReturnObj(y), an equivalent of lambda: 7
...     _1=Repeater(3),
...     y=lambda: next(counter),  # will return 0, then 1, then 2,...
...     z=lambda x, y: x * y),
...     1
... )
>>> list(f())
[{'x': 7, 'y': 0, 'z': 0}, {'x': 7, 'y': 1, 'z': 7}, {'x': 7, 'y': 2, 'z': 14}]


"""

from slink.seed_functions import (
    RandomCategoricalGenerator,
    RandomStringGenerator,
    RandomDictGenerator,
    RandomGenerator,
    rand_string,
)

from slink.sequences import (
    IterativeDictProcessing,
    Repeater,
    DictChain,
    dict_generator,
    mk_monotone_sequence,
)

from slink.util import (
    GetFromIter,
    select_fields,
)

# External tools that are useful to slink
from lined import CommandIter



"""Utils"""
from dataclasses import dataclass
from typing import Union, Iterable, Iterator, Optional
from functools import partial


field_is_missing = type('field_is_missing', (), {})()


def _select_fields(d: dict, fields, handle_missing: Optional[str] = None, default=None):
    """Helper of select_fields; selects fields from a single dict,
    with control on what to do if the field is missing.

    >>> d = {'a': 1, 'b': 2, 'c': 3}
    >>> dict(_select_fields(d, ['c', 'a']))
    {'c': 3, 'a': 1}
    >>> dict(_select_fields(d, ['c', 'd', 'a'], handle_missing='default'))
    {'c': 3, 'd': None, 'a': 1}
    """
    if not handle_missing:
        for field in fields:
            yield field, d[field]
    elif handle_missing == 'skip':
        for field in fields:
            val = d.get(field, field_is_missing)
            if val is not field_is_missing:
                yield field, d[field]
    elif handle_missing == 'default':
        for field in fields:
            yield field, d.get(field, default)
    else:
        raise ValueError(f'Unknown handle_missing value')


def select_fields(
    iterable_of_dicts: Iterable[dict],
    fields,
    handle_missing: Optional[str] = None,
    default=None,
):
    """Get an iterable of "sub-dicts". That is, dicts with only specified fields.
    Further, the  order of the fields will be as specified in the `fields` argument.

    By default `handle_missing=False`; any missing fields will raise a KeyError.
    To fallback to a default in this case, set `handle_missing=False` and
    optionally specify the desired default.

    >>> iterable_of_dicts = [{'a': 1, 'b': 2, 'c': 3}, {'a': 10, 'b': 20}]
    >>> list(select_fields(iterable_of_dicts, fields=['c', 'a'], handle_missing='skip'))
    [{'c': 3, 'a': 1}, {'a': 10}]
    >>> list(select_fields(
    ... iterable_of_dicts, fields=['c', 'a'], handle_missing='default', default=999)
    ... )
    [{'c': 3, 'a': 1}, {'c': 999, 'a': 10}]
    """
    selector = partial(
        _select_fields, fields=fields, handle_missing=handle_missing, default=default
    )
    return map(dict, map(selector, iterable_of_dicts))


@dataclass
class GetFromIter:
    """From an iterat(or)(able), get a function that iteratively returns the next item.

    Might become deprecated in the future because a very slightly more
    convenient equivalent to doing "manually":

    ```
    partial(next, iterator)
    # or
    partial(next, iter(iterable))
    ```

    """

    iterator: Union[Iterable, Iterator]

    def __post_init__(self):
        if not isinstance(self.iterator, Iterator):
            self.iterator = iter(self.iterator)

    def __call__(self):
        return next(self.iterator)



"""Examples and recipes"""


import itertools
from functools import partial

import numpy as np
from lined import iterize, Line, LineParametrized

# ---------------------------------------------------------------------------------------
# simple categorical map
cat_map = {'a': [1, 2, 3], 'b': [4, 5, 6]}
get_list_for_cat = cat_map.__getitem__

# to make the sequence function from this item func, you can do:
iterized_get_list_for_cat = iterize(get_list_for_cat)
assert list(iterized_get_list_for_cat('abaa')) == [
    [1, 2, 3],
    [4, 5, 6],
    [1, 2, 3],
    [1, 2, 3],
]

# ---------------------------------------------------------------------------------------
# simple categorical map with seed generator
from slink.seed_functions import RandomCategoricalGenerator


def get_list_for_cat(category):
    return cat_map[category]


seed_gen = RandomCategoricalGenerator(categories='ab')

g = Line(seed_gen, iterize(get_list_for_cat), list, lambda x: x)
g(n=5)
# Example: [[4, 5, 6], [1, 2, 3], [1, 2, 3], [4, 5, 6], [4, 5, 6]]

# ---------------------------------------------------------------------------------------
# categorical seeds generating segments from normal distributions
cat_2_parms_map = {
    'a': dict(loc=5, scale=0.5, size=3),
    'b': dict(loc=10, scale=0.3, size=2),
}
get_params_for_cat = cat_2_parms_map.__getitem__
call_normal_rand_on_params = lambda params: np.random.normal(**params)

_cat_based_chunk_gen = Line(
    get_params_for_cat,  # get some params for a category
    call_normal_rand_on_params,  # draw from a normal distribution with those params
    iterize(int),  # cast the result to an int
    list,  # make the arrays into lists (because it's nicer for display)
)
cat_based_chunk_gen = iterize(_cat_based_chunk_gen)
list(cat_based_chunk_gen('abaa'))
# Example: [[4, 5, 5], [9, 10], [5, 4, 4], [4, 5, 4]]

# ---------------------------------------------------------------------------------------
# _cat_based_chunk_gen.dot_digraph() doesn't work in the above.
# To make that work, need to make get_params_for_cat a function with signature
def get_params_for_cat(category):
    return cat_2_parms_map[category]


def call_normal_rand_on_params(params):
    return np.random.normal(**params)


_cat_based_chunk_gen = Line(
    get_params_for_cat,  # get some params for a category
    call_normal_rand_on_params,  # draw from a normal distribution with those params
    iterize(int),  # cast the result to an int
    list,  # make the arrays into lists (because it's nicer for display)
)
cat_based_chunk_gen = iterize(_cat_based_chunk_gen)
list(cat_based_chunk_gen('abaa'))
# Example: [[4, 5, 4], [10, 9], [5, 5, 4], [5, 5, 5]]

# ---------------------------------------------------------------------------------------
# TODO: Replace norm_offset and norm_scale by a num normalizing function
# TODO: Give control over spectr normalization
def get_spectr_from_number(num, spectr, norm_offset, norm_scale):
    """
    This function is meant to be curried so that only num is left.
    - spectr is meant to be fixed for a given dimension of data (where the nums are
    coming from)
    -
    """
    spectr = np.array(spectr)
    normalized_num = (num - norm_offset) / norm_scale
    original_spectr_sum = sum(spectr)
    spectr = spectr + normalized_num  # shift spectr by normalized_num
    factor_to_conserve_sum = original_spectr_sum / sum(spectr)
    return spectr * factor_to_conserve_sum


list_of_ints = Line(iterize(int), list,)

lists_of_list_of_ints = Line(iterize(list_of_ints), list,)


_num_based_wf_gen = LineParametrized(
    (
        'get_spectr',
        partial(
            get_spectr_from_number, spectr=[400, 900, 50], norm_offset=2, norm_scale=10
        ),
    ),
    ('wf_to_spectr', np.fft.rfft),
    ('normalize_rfft', lambda rfft_wf: np.abs(rfft_wf)),
)

num_based_wf_gen = iterize(_num_based_wf_gen)
list_of_ints(_num_based_wf_gen(10000))
# Example: [1350, 229]

num_based_wf_gen = iterize(_num_based_wf_gen)
lists_of_list_of_ints(num_based_wf_gen([4, 10000, 70]))
# Example: [[1350, 739], [1350, 229], [1350, 728]]

# ---------------------------------------------------------------------------------------



import pytest
import functools, random
from i2 import Sig
from slink.sequences import (
    ReturnObj,
    _call_from_dict,
    IterativeDictProcessing,
    Repeater,
    flatten_generators_recursively,
    ReturnObj,
    _prepare_formula_dict,
    _prepare_formulas,
    dict_generator,
    DictChain,
    mk_monotone_sequence,
)


def test_dummy():
    assert True


def test_call_from_dict():
    d = {'x': 3, 'multiply_by': 2, 'add_to': 4}

    def func(x, multiply_by, add_to):
        return x * multiply_by + add_to

    s = Sig(func)
    result = _call_from_dict(d, func, s)
    expected = 10
    assert result == expected


def test_IterativeDictProcessing():
    f = IterativeDictProcessing(
        phase=lambda session: session * 10,
        something_dependent=lambda session, phase: session + phase,
        something_independent=lambda: 'hi',
    )
    assert f({'session': 2}) == {
        'session': 2,
        'phase': 20,
        'something_dependent': 22,
        'something_independent': 'hi',
    }
    g = IterativeDictProcessing(
        session=functools.partial(random.uniform, 2, 9),
        block=lambda session: session * 10,
        something_dependent=lambda session, block: session + block,
        something_independent=lambda: random.choice([True, False, None]),
    )
    result = g()
    assert result['something_dependent'] == result['session'] + result['block']


def test_Repeater():
    rep = Repeater(3)
    assert list(rep('hello')) == ['hello', 'hello', 'hello']


def test_flatten_generators():
    def mk_gen(n):
        return (i * i for i in range(1, n))

    g = (mk_gen(i) for i in range(1, 5))
    assert list(flatten_generators_recursively(g)) == [1, 1, 4, 1, 4, 9]


def test_ReturnObj():
    t = ReturnObj('hey')
    assert t() == 'hey'


def test_prepare_formula_dict():
    message = 'hey'
    d = {'key': message}
    d_prep = _prepare_formula_dict(d)
    for k, v in d_prep:
        assert v() == message


def test_prepare_formula_dict():
    formulas = [{'key': 'word'}]
    for k, item in _prepare_formulas(formulas):
        assert ('key', item()) == ('key', 'word')


def test_dict_generator():
    import itertools

    counter = itertools.count()
    f = dict_generator(
        dict(
            x=7,  # will be replaced with ReturnObj(y), an equivalent of lambda: 7
            _1=Repeater(3),
            y=lambda: next(counter),  # will return 0, then 1, then 2,
            z=lambda x, y: x * y,
        ),
        1,
    )
    assert list(f()) == [
        {'x': 7, 'y': 0, 'z': 0},
        {'x': 7, 'y': 1, 'z': 7},
        {'x': 7, 'y': 2, 'z': 14},
    ]

    counter = itertools.count()
    f = dict_generator(
        dict(
            x=7,  # will be replaced with ReturnObj(y), an equivalent of lambda: 7
            _1=Repeater(3),
            y=lambda: next(counter),  # will return 3, then 4, then 5,
            z=lambda x, y: x * y,
        ),
        2,  # will be replaced with Repeater(2)
    )
    t = list(f())
    assert t == [
        {'x': 7, 'y': 0, 'z': 0},
        {'x': 7, 'y': 0, 'z': 0},
        {'x': 7, 'y': 1, 'z': 7},
        {'x': 7, 'y': 1, 'z': 7},
        {'x': 7, 'y': 2, 'z': 14},
        {'x': 7, 'y': 2, 'z': 14},
    ]


def test_DictChain():
    import itertools

    counter = itertools.count()

    f = DictChain(x=lambda: 7, _1=Repeater(3),)
    assert list(f()) == [{'x': 7}, {'x': 7}, {'x': 7}]
    f = DictChain(
        x=lambda: 7,
        _1=Repeater(3),
        y=lambda: next(counter),  # will return 0, then 1, then 2,
    )
    assert list(f()) == [{'x': 7, 'y': 0}, {'x': 7, 'y': 1}, {'x': 7, 'y': 2}]
    f = DictChain(
        x=lambda: 7,
        _1=Repeater(3),
        y=lambda: next(counter),  # will return 3, then 4, then 5,
        z=lambda x, y: x * y,
    )
    assert list(f()) == [
        {'x': 7, 'y': 3, 'z': 21},
        {'x': 7, 'y': 4, 'z': 28},
        {'x': 7, 'y': 5, 'z': 35},
    ]
    f = DictChain(
        x=lambda: 7,
        _1=Repeater(3),
        y=lambda: next(counter),  # will return 6, then 7, then 8,
        z=lambda x, y: x * y,
        _2=Repeater(2),
    )
    assert list(f()) == [
        {'x': 7, 'y': 6, 'z': 42},
        {'x': 7, 'y': 6, 'z': 42},
        {'x': 7, 'y': 7, 'z': 49},
        {'x': 7, 'y': 7, 'z': 49},
        {'x': 7, 'y': 8, 'z': 56},
        {'x': 7, 'y': 8, 'z': 56},
    ]


def test_mk_monotone_sequence():
    from lined import Pipe
    import itertools

    get_slice = Pipe(itertools.islice, list)

    def func(begin, end, start=0):
        return 42

    it = mk_monotone_sequence(func, 10, 100, start=7)
    t = get_slice(it, 3)
    assert len(t) == 3
    assert t == [7, 49, 91]
